/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach.bayernid.proxy;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import java.util.concurrent.atomic.AtomicReference;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;
import org.springframework.test.util.ReflectionTestUtils;

import com.google.protobuf.ByteString;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.Attachment;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessage.BayernIdMessageBuilder;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessageMapper;
import de.ozgcloud.nachrichten.postfach.bayernid.proxy.message.BayernIdMessageTestFactory;
import io.grpc.stub.StreamObserver;
import lombok.SneakyThrows;

class UploadStreamObserverTest {

	@Spy
	@InjectMocks
	private UploadStreamObserver uploadStreamObserver;

	@Mock
	private StreamObserver<GrpcSendBayernIdMessageResponse> responseObserver;
	@Mock
	private BayernIdProxyService proxyService;
	@Mock
	private BayernIdMessageMapper messageMapper;
	@Mock
	private BayernIdMessageBuilder bayernIdMessageBuilder;
	@Mock
	private AtomicReference<BayernIdMessageBuilder> messageBuilderAtomicReference;

	@Nested
	class TestOnNext {

		@Test
		void shouldCallReadMetadata() {
			var messageMetadata = GrpcBayernIdMessageMetadataTestFactory.create();
			var sendRequest = GrpcSendBayernIdMessageRequest.newBuilder().setMessageMetadata(messageMetadata).build();

			uploadStreamObserver.onNext(sendRequest);

			verify(uploadStreamObserver).readMessageMetadata(messageMetadata);
			verify(uploadStreamObserver, never()).readAttachments(any());
		}

		@Test
		void shouldCallReadAttachment() {
			var attachments = GrpcAttachments.newBuilder().setAttachmentMetadata(GrpcAttachmentMetadataTestFactory.create()).build();
			var sendRequest = GrpcSendBayernIdMessageRequest.newBuilder().setAttachments(attachments).build();
			doNothing().when(uploadStreamObserver).readAttachments(any());

			uploadStreamObserver.onNext(sendRequest);

			verify(uploadStreamObserver).readAttachments(attachments);
			verify(uploadStreamObserver, never()).readMessageMetadata(any());
		}
	}

	@Nested
	class TestReadMessageMetadata {

		private GrpcBayernIdMessageMetadata messageMetadata = GrpcBayernIdMessageMetadataTestFactory.create();

		@BeforeEach
		void setup() {
			ReflectionTestUtils.setField(uploadStreamObserver, "messageBuilder", messageBuilderAtomicReference);
		}

		@Test
		void shouldCallMessageMapper() {
			uploadStreamObserver.readMessageMetadata(messageMetadata);

			verify(messageMapper).toBayernIdMessage(messageMetadata);
		}

		@Test
		void shouldSetMessageBuilder() {
			when(messageMapper.toBayernIdMessage(any())).thenReturn(bayernIdMessageBuilder);

			uploadStreamObserver.readMessageMetadata(messageMetadata);

			verify(messageBuilderAtomicReference).set(bayernIdMessageBuilder);
		}

	}

	@Nested
	class TestReadAttachments {

		@Test
		void shouldCallAddContainer() {
			var attachmentMetadata = GrpcAttachmentMetadataTestFactory.create();
			doNothing().when(uploadStreamObserver).addDataContainer(any());

			uploadStreamObserver.readAttachments(GrpcAttachments.newBuilder().setAttachmentMetadata(attachmentMetadata).build());

			verify(uploadStreamObserver).addDataContainer(attachmentMetadata);
			verify(uploadStreamObserver, never()).writeAttachment(any());
		}

		@Test
		void shouldCallWriteAttachment() {
			var content = ByteString.copyFromUtf8("attachment");
			var attachment = GrpcAttachments.newBuilder().setContent(content).build();
			doNothing().when(uploadStreamObserver).writeAttachment(any());

			uploadStreamObserver.readAttachments(attachment);

			verify(uploadStreamObserver).writeAttachment(content);
			verify(uploadStreamObserver, never()).addDataContainer(any());
		}

		@Nested
		class TestAddDataContainer {

			private GrpcAttachmentMetadata grpcAttachmentMetadata = GrpcAttachmentMetadataTestFactory.create();

			@Mock
			private Attachment attachment;

			@Mock
			private AtomicReference<BayernIdMessageBuilder> atomicMessageBuilder;

			@BeforeEach
			void setup() {
				when(atomicMessageBuilder.get()).thenReturn(bayernIdMessageBuilder);
				ReflectionTestUtils.setField(uploadStreamObserver, "messageBuilder", atomicMessageBuilder);
			}

			@Test
			void shouldCallBuildAttachment() {
				uploadStreamObserver.addDataContainer(grpcAttachmentMetadata);

				verify(uploadStreamObserver).buildAttachment(grpcAttachmentMetadata);
			}

			@Test
			void shouldAddAttachment() {
				doReturn(attachment).when(uploadStreamObserver).buildAttachment(any());

				uploadStreamObserver.addDataContainer(grpcAttachmentMetadata);

				verify(bayernIdMessageBuilder).attachment(attachment);
			}
		}

		@Nested
		class TestBuildAttachment {

			private GrpcAttachmentMetadata grpcAttachmentMetadata = GrpcAttachmentMetadataTestFactory.create();

			@Mock
			private InputStream tmpFileInputStream;

			@Nested
			class TestBuildSuccessfully {
				@SneakyThrows
				@BeforeEach
				void setup() {
					doReturn(tmpFileInputStream).when(uploadStreamObserver).createTemporallyFile();
				}

				@SneakyThrows
				@Test
				void shouldCallCreateTemporallyFile() {
					uploadStreamObserver.buildAttachment(grpcAttachmentMetadata);

					verify(uploadStreamObserver).createTemporallyFile();
				}

				@Test
				void shouldSetTemporallyFile() {
					var result = uploadStreamObserver.buildAttachment(grpcAttachmentMetadata);

					assertThat(result.getContent()).isEqualTo(tmpFileInputStream);
				}

				@Test
				void shouldSetAttachmentName() {
					var result = uploadStreamObserver.buildAttachment(grpcAttachmentMetadata);

					assertThat(result.getName()).isEqualTo(GrpcAttachmentMetadataTestFactory.FILE_NAME);
				}

				@Test
				void shouldSetAttachmentType() {
					var result = uploadStreamObserver.buildAttachment(grpcAttachmentMetadata);

					assertThat(result.getType()).isEqualTo(GrpcAttachmentMetadataTestFactory.FILE_TYPE);
				}
			}

			@SneakyThrows
			@Test
			void shouldThrowTechnicalException() {
				doThrow(IOException.class).when(uploadStreamObserver).createTemporallyFile();

				assertThrows(TechnicalException.class, () -> uploadStreamObserver.buildAttachment(grpcAttachmentMetadata));
			}
		}

		@Nested
		class TestCreateTemporallyFile {

			private static final Path TEMP_FILE_PATH = Path.of("tempFile");

			@Mock
			private OutputStream attachmentWriter;
			@Mock
			private InputStream tmpFileInputStream;

			private MockedStatic<TempFileUtils> mockTempFileUtils;

			@SneakyThrows
			@BeforeEach
			void init() {
				mockTempFileUtils = mockStatic(TempFileUtils.class);
				mockTempFileUtils.when(TempFileUtils::createTmpFile).thenReturn(TEMP_FILE_PATH);
				doReturn(tmpFileInputStream).when(uploadStreamObserver).newInputStream(any());
			}

			@AfterEach
			void close() {
				mockTempFileUtils.close();
			}

			@SneakyThrows
			@Test
			void shouldCallCloseTemporallyFile() {
				uploadStreamObserver.createTemporallyFile();

				verify(uploadStreamObserver).closeTemporallyWriter();
			}

			@SneakyThrows
			@Test
			void shouldCallCreateFile() {
				mockTempFileUtils.when(TempFileUtils::createTmpFile).thenReturn(TEMP_FILE_PATH);

				uploadStreamObserver.createTemporallyFile();

				mockTempFileUtils.verify(() -> TempFileUtils.createTmpFile());
			}

			@SneakyThrows
			@Test
			void shouldCallNewOutputStream() {
				uploadStreamObserver.createTemporallyFile();

				verify(uploadStreamObserver).newOutputStream(TEMP_FILE_PATH);
			}

			@SneakyThrows
			@Test
			void shouldInitializeOutputWriter() {
				doReturn(attachmentWriter).when(uploadStreamObserver).newOutputStream(any());

				uploadStreamObserver.createTemporallyFile();

				assertThat(getAttachmentWriter()).isEqualTo(attachmentWriter);
			}

			@SneakyThrows
			@Test
			void shouldCallNewInputStream() {
				uploadStreamObserver.createTemporallyFile();

				verify(uploadStreamObserver).newInputStream(TEMP_FILE_PATH);
			}

			@SneakyThrows
			@Test
			void shouldReturnInputStream() {
				var result = uploadStreamObserver.createTemporallyFile();

				assertThat(result).isEqualTo(tmpFileInputStream);
			}

			private OutputStream getAttachmentWriter() {
				return (OutputStream) ReflectionTestUtils.getField(uploadStreamObserver, "attachmentWriter");
			}
		}
	}

	@Nested
	class TestOnError {

		@Test
		void shouldCallCloseTemporallyWriter() {
			try {
				uploadStreamObserver.onError(new RuntimeException());
			} catch (TechnicalException e) {
				// expected
			}

			verify(uploadStreamObserver).closeTemporallyWriter();
		}

		@Test
		void shouldTrowTechnicalException() {
			var cause = new RuntimeException();

			var thrownException = assertThrows(TechnicalException.class, () -> uploadStreamObserver.onError(cause));

			assertThat(thrownException.getCause()).isEqualTo(cause);
		}
	}

	@Nested
	class TestOnComplete {

		@BeforeEach
		void setup() {
			when(messageBuilderAtomicReference.get()).thenReturn(bayernIdMessageBuilder);
			ReflectionTestUtils.setField(uploadStreamObserver, "messageBuilder", messageBuilderAtomicReference);
		}

		@Test
		void shouldCallCloseTemporallyWriter() {
			uploadStreamObserver.onCompleted();

			verify(uploadStreamObserver).closeTemporallyWriter();
		}

		@Test
		void shouldCallSendBayernIdMessage() {
			var bayernIdMessage = BayernIdMessageTestFactory.create();
			when(bayernIdMessageBuilder.build()).thenReturn(bayernIdMessage);

			uploadStreamObserver.onCompleted();

			verify(proxyService).sendPostfachNachricht(bayernIdMessage);
		}

		@Test
		void shouldCallMessageMapper() {
			var bayernIdResponse = BayernIdResponseTestFactory.create();
			when(proxyService.sendPostfachNachricht(any())).thenReturn(bayernIdResponse);

			uploadStreamObserver.onCompleted();

			verify(messageMapper).fromBayernIdResponse(bayernIdResponse);
		}

		@Test
		void shouldCallResponseObserver() {
			var grpcSendResponse = GrpcSendBayernIdMessageResponse.newBuilder().build();
			when(messageMapper.fromBayernIdResponse(any())).thenReturn(grpcSendResponse);

			uploadStreamObserver.onCompleted();

			verify(responseObserver).onNext(grpcSendResponse);
		}

		@Test
		void shouldCallSendResponse() {
			uploadStreamObserver.onCompleted();

			verify(responseObserver).onNext(any());
		}

		@Test
		void shouldCallComplete() {
			uploadStreamObserver.onCompleted();

			verify(responseObserver).onCompleted();
		}

		@Test
		void shouldThrowTechnicalException() {
			var cause = new RuntimeException();

			doThrow(cause).when(proxyService).sendPostfachNachricht(any());

			var receivedException = assertThrows(TechnicalException.class, () -> uploadStreamObserver.onCompleted());
			assertThat(receivedException.getCause()).isEqualTo(cause);
		}
	}
}